
#include <inttypes.h> // intX_t
#include <sys/types.h> // X_intX_t;
#include <string.h> // memcpy

#include <stdio.h> // printf

#include <stdlib.h> // malloc / free

uint8_t raw [] = { 0x00, 0xa0, 0x00, 0x14, 0x00, 0x00, 0x85, 0x56,
		  0x00, 0x14, 0xbe, 0x28, 0x00, 0x5d, 0x00, 0x14,
		  0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

uint8_t *AcctableBuf;

#ifdef AVISION

typedef uint8_t  BYTE;
typedef uint16_t WORD;
typedef uint32_t DWORD;


namespace CUtilities {
  WORD SwapWORD(WORD w) {
    return (w & 0xFF) << 8 | (w >> 8);
  }
}
#endif

void debug_print_hex_raw (const char* msg, const uint8_t* data, size_t count)
{
  int address = 0;
  char text [16*3+1];
  
  printf (msg);
  
  while (count) {
    char* t = text;
    int i = 0;
    while (i < 16 && count) {
      t += sprintf (t, "%02x ", *data++);
      count--; i++;
    }
    *--t = 0;
    
    printf ("  [%08x] %s\n", address, text);
    address += 16;
  }
}

#ifdef AVISION

typedef struct {
  WORD totalSteps;
  WORD stableSteps;
  DWORD tableUnits;
  DWORD baseUnits;
  WORD startSpeed;
  WORD targetSpeed;
  BYTE ability;
  BYTE tableCount;
  BYTE reserved[6];
} __attribute__((packed)) ACCELERATE_INFO, *LPACCELERATE_INFO;


//acceleration info
ACCELERATE_INFO accelInfo;
BYTE  *accelTable;
BYTE  tableIndex=0;
DWORD dwDummy;

WORD ul_TotalBase, ul_ClockUnit, ui_StartSpeed, ui_TargetSpeed;

WORD dwErrorCode;

#define AV_ALLOCATE_MEMORY_FAILED 1

DWORD li_TotalCount(WORD ui_TotalStep)
{
  DWORD li_Total ;
  WORD i_TabIndex ;
  BYTE              *buf = AcctableBuf;

  for (li_Total=0,i_TabIndex=0 ; i_TabIndex<ui_TotalStep ;
       i_TabIndex++)
    li_Total += buf[i_TabIndex] ;
  return (li_Total) ;
}

void FixTable(WORD ui_TotalStep)
{
  WORD              ui_TabIndex,ui_TempCount ;
  DWORD             ul_AddCount ;
  BYTE              *buf = AcctableBuf;

  ui_TabIndex=0 ;
  if (((li_TotalCount(ui_TotalStep)*ul_ClockUnit) % ul_TotalBase) == 0)
    ul_AddCount = 0 ;
  else
    ul_AddCount =
      (ul_TotalBase-((li_TotalCount(ui_TotalStep)*ul_ClockUnit)%ul_TotalBase))/ul_ClockUnit
      ;
  // ul_AddCount should not bigger than 255
  if(ul_AddCount > 255)
    ul_AddCount = 255;
  while ((ui_TabIndex<(ui_TotalStep-1)) && (ul_AddCount!=0))
    {
      ui_TempCount = 255-buf[ui_TabIndex] ;
      if (ui_TempCount > ul_AddCount)
	{
	  buf[ui_TabIndex] += (BYTE)ul_AddCount ;
	  ul_AddCount = 0 ;
	}
      else
	{
	  buf[ui_TabIndex] += ui_TempCount ;
	  ul_AddCount -= ui_TempCount ;
	}
      ui_TabIndex++ ;
    }
}

WORD ui_FillTable(float f_AccRate)
{
  float             f_NowCount ;
  WORD              ui_TabCount ;
  WORD              i_NowCount =ui_StartSpeed ;
  BYTE              *buf = AcctableBuf;


  ui_TabCount = 1 ;
  f_NowCount = ui_StartSpeed ;
  buf[0] = (BYTE)(ui_StartSpeed) ;

  while (i_NowCount != ui_TargetSpeed)
    {
      f_NowCount = f_NowCount - (f_NowCount -
				 ui_TargetSpeed)*f_AccRate ;
      i_NowCount = (WORD)(f_NowCount+0.5) ;
      buf[ui_TabCount]=(BYTE)(i_NowCount) ;
      ui_TabCount++ ;
      if(ui_TabCount >= 999)
	{
	  // Buffer is around 1000 bytes, can't extend it.
	  break;
	}
    } ;
  return  ((WORD)ui_TabCount) ;
}

BYTE* GetAccTable(LPACCELERATE_INFO pAccelInfo)
{
  float                         f_UpLim,f_LowLim,f_Mid ;
  WORD                          ui_AccStep;
  WORD                          ui_TotalStep;
  WORD                          i;
  BYTE              *buf=AcctableBuf;




  memset(buf,0,1000);
  ui_AccStep = pAccelInfo->totalSteps - pAccelInfo->stableSteps + 1;
  ul_ClockUnit = pAccelInfo->tableUnits;
  ul_TotalBase = pAccelInfo->baseUnits;
  ui_StartSpeed = pAccelInfo->startSpeed;
  ui_TargetSpeed = pAccelInfo->targetSpeed;

  f_UpLim=1 ;
  f_LowLim=(float)0.001 ;
  while ((f_UpLim-f_LowLim) > 0.0001)
    {
      f_Mid = (f_UpLim+f_LowLim)/2 ;
      ui_TotalStep = ui_FillTable(f_Mid) ;


      if (ui_TotalStep == ui_AccStep)
	break ;
      if (ui_TotalStep > ui_AccStep)
	f_LowLim = f_Mid ;
      else
	f_UpLim = f_Mid ;
    }
  for(i=ui_AccStep-1;i<pAccelInfo->totalSteps;i++)
    {
      buf[i+1] = buf[i];
    }
  debug_print_hex_raw ("first:\n",buf, accelInfo.totalSteps);
  FixTable(pAccelInfo->totalSteps) ;
  debug_print_hex_raw ("fixed:\n",buf, accelInfo.totalSteps);

  for(i=0;i<pAccelInfo->totalSteps;i++)
    {
      buf[i] --;
    }
  debug_print_hex_raw ("final:\n",buf, accelInfo.totalSteps);
  return buf;

}
#endif

int main ()
{
#ifdef AVISION
  AcctableBuf=new BYTE[1000];
  if(NULL == AcctableBuf)
    {
      dwErrorCode=AV_ALLOCATE_MEMORY_FAILED;
      return false;
    }

  //----------------------------Control accelerate
  WORD dwLen = sizeof(accelInfo);
  do
    {
      /*
	dwErrorCode=MyIO->SetReadCmd(0x6c,&dwLen,0,(LPBYTE)&accelInfo,&dwDummy);
	if(0 != dwErrorCode)
	{
	//CUtilities::dwMyErrorCode = 250;
	//return false;
	break;
	}
      */
     
      memcpy (&accelInfo, raw, sizeof (accelInfo));
      
      if(accelInfo.tableCount == 0)// Firmware don't need diver to download accelerated table.
	break;

      // Read information for accelerated table  calculation.
      accelInfo.totalSteps =
	CUtilities::SwapWORD(accelInfo.totalSteps);
      accelInfo.stableSteps =
	CUtilities::SwapWORD(accelInfo.stableSteps);
      accelInfo.tableUnits =
	CUtilities::SwapWORD(accelInfo.tableUnits);
      accelInfo.baseUnits =
	CUtilities::SwapWORD(accelInfo.baseUnits);
      accelInfo.startSpeed =
	CUtilities::SwapWORD(accelInfo.startSpeed);
      accelInfo.targetSpeed =
	CUtilities::SwapWORD(accelInfo.targetSpeed);
      if(accelInfo.targetSpeed > accelInfo.startSpeed ||
	 accelInfo.targetSpeed == 0 ||
	 accelInfo.totalSteps <=
	 accelInfo.stableSteps)
	{
	  delete [] AcctableBuf;
	  return false;
	  break;
	}
      if(0 != accelInfo.ability || 0 != dwErrorCode)
	{
	  // Something wrong in here.  All bits should be turnned off
	  delete [] AcctableBuf;
	  return false;
	  break;

	}
      else
	{
	  accelTable = GetAccTable(&accelInfo);
	  /*
	    dwErrorCode=MyIO->SetSendCmd(0x6c,tableIndex,accelInfo.totalSteps,AcctableBuf);
	    if(0 != dwErrorCode)
	    {
	    delete [] AcctableBuf;
	    CUtilities::dwMyErrorCode =
	    dwErrorCode;
	    return false;
	    break;
	    }
	  */
	  //SetSendCmd( 0x6c, tableIndex, accelInfo.totalSteps ,gAcctableBuf );
	}
      tableIndex ++;

    } while(accelInfo.tableCount > tableIndex);


  delete [] AcctableBuf;
  #endif
  // ---------------------------------------------------
  {
    typedef struct acceleration_info
    {
      u_int16_t total_steps;
      u_int16_t stable_steps;
      u_int32_t table_units;
      u_int32_t base_units;
      u_int16_t start_speed;
      u_int16_t target_speed;
      u_int8_t ability;
      u_int8_t table_count;
      u_int8_t reserved[6];
    } acceleration_info;
    
    int i, table = 0;
    acceleration_info accel_info;
    u_int8_t* table_data;
    
    u_int8_t* result = raw;
    acceleration_info* info = &accel_info;
    
    printf ("Computing table %d\n", table);
    
#define get_double(var) ((*var << 8) + *(var + 1))
#define get_quad(var)   ((*var << 24) +				\
                         (*(var + 1) << 16) +			\
                         (*(var + 2) << 8) + *(var + 3))
    
    info->total_steps = get_double (&(result[0]));
    info->stable_steps = get_double (&(result[2]));
    info->table_units = get_quad (&(result[4]));
    info->base_units = get_double (&(result[8]));
    info->start_speed = get_double (&(result[12]));
    info->target_speed = get_double (&(result[14]));
    info->ability = result[16];
    info->table_count = result[17];

    if (accel_info.table_count == 0) {
      printf ("send_acceleration_table: devices does not need tables\n");
      return 1;
    }
    
    if(accel_info.target_speed > accel_info.start_speed ||
       accel_info.target_speed == 0 ||
       accel_info.total_steps <= accel_info.stable_steps) {
      printf ("send_acceleration_table: tables does not look right.\n");
      return 2;
    }
    
    if(accel_info.ability != 0) {
      printf ("send_acceleration_table: ability non-zero - code needs changes\n");
      return 3;
    }
    do {
      /* so far I assume we have one byte tables as used in the C6 ASSIC ... */
      table_data = (u_int8_t*) malloc (accel_info.total_steps);
      float low_lim = 0.001;
      float up_lim  = 1.0;

      u_int16_t accel_steps = accel_info.total_steps - accel_info.stable_steps + 1;
      
      /* acceleration ramp */
      while ((up_lim - low_lim) > 0.0001)
	{
	  float mid = (up_lim + low_lim) / 2; /* accel rate */
	
	  u_int16_t now_count = accel_info.start_speed;
	
	  u_int16_t i = 0;
	
	  float now_count_f = now_count;
	  table_data [i++] = (u_int8_t) accel_info.start_speed;

	  while (now_count != accel_info.target_speed)
	    {
	      now_count_f = now_count_f - (now_count_f -
					   accel_info.target_speed) * mid;
	      now_count = (u_int16_t)(now_count_f + 0.5);
	      table_data[i++] = (u_int8_t) now_count;
	    }
	
	
	  if (i == accel_steps)
	    break;
	  if (i > accel_steps)
	    low_lim = mid;
	  else
	    up_lim = mid;
	}
      
      /* fill stable steps */
      for (i = accel_steps; i < accel_info.total_steps; i++)
	{
	  table_data [i] = table_data [i-1];
	}
      
          
      debug_print_hex_raw ("first:\n",
			   table_data, accel_info.total_steps);
      
      /* maybe post fix-up */
      {
	int add_count;
	
	/* count total steps in table */
	int table_total = 0;
	for (i = 0; i < accel_info.total_steps; i++)
	  table_total += table_data [i];
	
	i = 0;
	if (((table_total * accel_info.table_units) % accel_info.base_units) == 0)
	  add_count = 0;
	else
	  add_count = (accel_info.base_units -
		       ((table_total*accel_info.table_units) % accel_info.base_units))
	    / accel_info.table_units;
	
	/* add_count should not be bigger than 255 */
	if (add_count > 255) {
	  printf ("send_acceleration_table: add_count limitted, was: %d\n", add_count);
	  add_count = 255;
	}
	for (i = 0; i < accel_info.total_steps - 1 && add_count > 0; i++)
	  {
	    u_int16_t temp_count = 255 - table_data [i];
	    temp_count = temp_count > add_count ? add_count : temp_count;
	    
	    table_data [i] += (u_int8_t) temp_count;
	    add_count -= temp_count;
	  }
      }
    
      debug_print_hex_raw ("fixed:\n",
			   table_data, accel_info.total_steps);
    
      /* decrease all by one ... */
      for (i = 0; i < accel_info.total_steps; i++) {
	table_data[i]--;
      }
    
      debug_print_hex_raw ("final:\n",
			   table_data, accel_info.total_steps);
    
    
      free (table_data); table_data = 0;
    
      table++;
    } while (table < accel_info.table_count);
  }
  
  return 0;
}
